/*******************************************************************************
 *
 * File Name: LRRCgameAI.cpp
 * Author: Richard Baxter
 * Project: Lego Rules CG Rounds Checker
 * Project Version: 1l3j 2-Sept-08
 * Project First Internal Release: 1aXx 18-Sept-05 (C)
 * Project Second Internal Release: 2aXx 02-April-06 (convert to C++)
 * Project Third Internal Release: 2b7d 26-Sept-06 (added sprites)
 * Project Fourth Internal Release: 2c4a 25-Feb-07 [rotation of models supported, copyCombatRelevantInheritedImmediateParentUnitDetailsIntoChildObject() no longer used]
 * Project Fifth Internal Release: 2d11a 3-June-07
 *
 * NB phase specific sprites are yet to be added to scene files based upon movement/combat outcomes
 * NB phase specific sprites are yet to be added to scene files based upon future movement/combat options
 *******************************************************************************/

#include "LRRCgameAI.h"
#include "LDparser.h"
#include "LRRCgameReferenceManipulation.h"
#include "XMLrulesClass.h"
#include "LRRCcombat.h"


#define CHAR_COMMA ','
#define CHAR_NEW_LINE '\n'






void parseSceneFileAndFillUnitLists(char sceneFileName[], UnitListClass * firstUnitInUnitList, int currentRound)
{
	bool result = true;

	Reference * initialReferenceInSceneFile = new Reference();
	Reference * topLevelReferenceInSceneFile = new Reference(sceneFileName, 1, true);	//The information in this object is not required or meaningful, but needs to be passed into the parseFile/parseReferenceList recursive function
	if(!parseFile(sceneFileName, initialReferenceInSceneFile, topLevelReferenceInSceneFile))
	{//file does not exist
		cout << "The file: " << sceneFileName << " does not exist in the directory" << endl;
		result = false;
	}

	fillUnitList(initialReferenceInSceneFile, firstUnitInUnitList, currentRound);

	delete initialReferenceInSceneFile;
	delete topLevelReferenceInSceneFile;
}





void fillUnitList(Reference * currentReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound)
{
	Reference * currentReference = currentReferenceInSceneFile;
	UnitListClass * currentUnitInUnitList = firstUnitInUnitGroup;
	while(currentReference->next != NULL)
	{
		if(currentReference->isSubModelReference)
		{
			//cout << "currentReference->name = " << currentReference->name << endl;

			Reference * topLevelReferenceInUnit = new Reference(currentReference->name, currentReference->colour, true);
			copyAllUnitDetails(topLevelReferenceInUnit->subModelDetails, currentReference->subModelDetails);
			searchSceneReferenceListAndDetermineTheDetailsOfAParticularUnitSubmodel(topLevelReferenceInUnit, currentReference->firstReferenceWithinSubModel, topLevelReferenceInUnit, true);
			/*
			//debug:
			RecordClass * currentReferenceRecordClass = topLevelReferenceInUnit->subModelDetails->recordOfUnitTypeDetails;
			while(currentReferenceRecordClass->next != NULL)
			{
				if(currentReferenceRecordClass->name == PERSON_HEAD_NAME)
				{
					cout << "topLevelReferenceInUnit->name = " << topLevelReferenceInUnit->name << endl;
					cout << "\t PERSON_HEAD_NAME topLevelReferenceInUnit->numberOfThisPartIDInTheUnit = " << currentReferenceRecordClass->numberOfThisPartIDInTheUnit << endl;
				}

				currentReferenceRecordClass = currentReferenceRecordClass->next;
			}
			*/
			determineUnitTypeAndMinSpeedOfUnitGroup(topLevelReferenceInUnit->subModelDetails);



			if(topLevelReferenceInUnit->subModelDetails->numPerson >= 1)
			{//add a unit to the unit list if it contains a single person

				//add a unit/group to the unit list

				currentUnitInUnitList->name = currentReference->name;
				currentUnitInUnitList->team = currentReference->colour;	//not currently used, as there is 1 UnitListClass created for every team
				currentUnitInUnitList->unitDetails = topLevelReferenceInUnit->subModelDetails;		//OR CHECK THIS; currentReference->subModelDetails or topLevelReferenceInUnit->subModelDetails??? it depends upon what information is to be recorded here
				currentUnitInUnitList->status = true;
				currentUnitInUnitList->roundSpawned = currentRound;

				UnitListClass * newUnitList = new UnitListClass();
				currentUnitInUnitList->next = newUnitList;

				if(topLevelReferenceInUnit->subModelDetails->numPerson == 1)
				{
					//do not go down further into reference tree if already reached a unit containing only a single person
				}
				else
				{
					//go down further into reference tree
					UnitListClass * newUnitList1 = new UnitListClass();
					currentUnitInUnitList->firstUnitInUnitGroup = newUnitList1;
					currentUnitInUnitList->isUnitGroup = true;
					fillUnitList(currentReference->firstReferenceWithinSubModel, currentUnitInUnitList->firstUnitInUnitGroup, currentRound);
				}

				currentUnitInUnitList=currentUnitInUnitList->next;
			}
			
			delete topLevelReferenceInUnit;
		}

		currentReference = currentReference->next;
	}
}





bool determineIfUnitGroupHasAliveUnits(UnitListClass * firstUnitInUnitGroup)
{
	bool foundAliveUnit = false;
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		if(currentUnitInList->status == true)
		{
			foundAliveUnit = true;
	
		}

		if(currentUnitInList->isUnitGroup)
		{
			if(determineIfUnitGroupHasAliveUnits(currentUnitInList->firstUnitInUnitGroup))
			{
				foundAliveUnit = true;
			}
		}

		currentUnitInList = currentUnitInList->next;
	}
	return foundAliveUnit;
}

bool checkAverageKillRatioForUnitGroup(UnitListClass * firstUnitInUnitGroup)
{
	bool result;
	
	int unitGroupTotalKillPoints = 0;
	int unitGroupTotalDeathPoints = 0;
	determineAverageKillRatioForUnitGroup(firstUnitInUnitGroup, &unitGroupTotalKillPoints, &unitGroupTotalDeathPoints);
	if(double(unitGroupTotalKillPoints)/double(unitGroupTotalDeathPoints) >= GAME_AI_NN_EXPERIENCE_RECORD_MIN_KILL_RATIO_REQUIREMENT)
	{
		result = true;
	}
	else
	{
		result = false;
	}

	return result;
}

void determineAverageKillRatioForUnitGroup(UnitListClass * firstUnitInUnitGroup, int * killPoints, int * deathPoints)
{
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		if(currentUnitInList->isUnitGroup)
		{
			determineAverageKillRatioForUnitGroup(currentUnitInList->firstUnitInUnitGroup, killPoints, deathPoints);
		}
		else
		{
			*killPoints = *killPoints + currentUnitInList->killPoints;
			if(currentUnitInList->status == false)
			{
				*deathPoints = *deathPoints + determineUnitWorthInPoints(currentUnitInList->unitDetails);
			}		
		}

		currentUnitInList = currentUnitInList->next;
	}
}

int determineUnitWorthInPoints(ModelDetails * unitDetails)
{
	int totalPoints;
	
	int movementPoints;
	int longDistancePoints;
	int closeCombatPoints;
	
	movementPoints = unitDetails->movementSpeed/GAME_UNIT_POINTS_MOVEMENT_RANGE_MODIFIER;
	longDistancePoints = unitDetails->longDistanceAttackTotal + unitDetails->longDistanceAttackBaseRange/GAME_UNIT_POINTS_LONG_DISTANCE_RANGE_MODIFIER;
	closeCombatPoints = unitDetails->closeCombatAttackTotal;
	
	totalPoints = (movementPoints + maxInt(longDistancePoints, closeCombatPoints));
	
	return totalPoints;	
}



void parseSceneFileAndUpdateUnitList(char sceneFileName[], UnitListClass * firstUnitInUnitList, int currentRound)
{
	bool result = true;

	Reference * initialReferenceInSceneFile = new Reference();
	Reference * topLevelReferenceInSceneFile = new Reference(sceneFileName, 1, true);	//The information in this object is not required or meaningful, but needs to be passed into the parseFile/parseReferenceList recursive function
	if(!parseFile(sceneFileName, initialReferenceInSceneFile, topLevelReferenceInSceneFile))
	{//file does not exist
		cout << "The file: " << sceneFileName << " does not exist in the directory" << endl;
		result = false;
	}

	updateUnitList(initialReferenceInSceneFile, firstUnitInUnitList, currentRound);
	
	delete initialReferenceInSceneFile;
	delete firstUnitInUnitList;
}

void updateUnitList(Reference * initialReferenceInSceneFile, UnitListClass * firstUnitInUnitList, int currentRound)	//required after every round
{
	updateUnitListWithNewUnits(initialReferenceInSceneFile, firstUnitInUnitList, currentRound);

	updateUnitListWithDeadUnits(initialReferenceInSceneFile, firstUnitInUnitList, currentRound);

}

void updateUnitListWithNewUnits(Reference * currentReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound)
{
	Reference * currentReference = currentReferenceInSceneFile;
	while(currentReference->next != NULL)
	{
		if(currentReference->isSubModelReference)
		{
			//add a unit to the unit list if it contains a single person

			bool foundUnitInList = false;
			UnitListClass * foundUnit;
			foundUnit = searchUnitListForUnitAndIfNotThereAddToList(firstUnitInUnitGroup, currentReference, currentRound, true, &foundUnitInList);

			//go down further into reference tree
			if(foundUnitInList)
			{
				if(foundUnit->isUnitGroup)
				{//do nothing

				}
				else
				{
					foundUnit->isUnitGroup = true;
					UnitListClass * newUnit = new UnitListClass();
					foundUnit->firstUnitInUnitGroup = newUnit;
				}
			}
			else
			{
				foundUnit->isUnitGroup = true;
				UnitListClass * newUnit = new UnitListClass();
				foundUnit->firstUnitInUnitGroup = newUnit;

			}
			updateUnitListWithNewUnits(currentReference->firstReferenceWithinSubModel, foundUnit->firstUnitInUnitGroup, currentRound);
		}

		currentReference = currentReference->next;
	}
}

void updateUnitListWithDeadUnits(Reference * initialReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound)
{
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		bool unitIDFound = false;
		bool result = true;

		Reference * foundReference = searchReferenceListFindReference(initialReferenceInSceneFile, currentUnitInList->name, currentUnitInList->team, &unitIDFound, &result);
		if(!unitIDFound)
		{
			currentUnitInList->status = false;
			currentUnitInList->roundRemoved = currentRound;
		}

		if(currentUnitInList->isUnitGroup)
		{
			updateUnitListWithDeadUnits(initialReferenceInSceneFile, currentUnitInList->firstUnitInUnitGroup, currentRound);
		}

		currentUnitInList = currentUnitInList->next;
	}
}

UnitListClass * searchUnitListForUnitAndIfNotThereAddToList(UnitListClass * firstUnitInUnitGroup, Reference * unitReferenceInSceneFile, int currentRound, bool topLevelInTree, bool * foundUnitInList)
{
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;
	UnitListClass * foundUnit;

	while((currentUnitInList->next != NULL) && (!(*foundUnitInList)))
	{
		if((currentUnitInList->name == unitReferenceInSceneFile->name) && (currentUnitInList->team == unitReferenceInSceneFile->colour))
		{
			*foundUnitInList = true;
			foundUnit = currentUnitInList;
		}
		else
		{
			if(currentUnitInList->isUnitGroup)
			{
				UnitListClass * tempUnit;
				tempUnit = searchUnitListForUnitAndIfNotThereAddToList(currentUnitInList->firstUnitInUnitGroup, unitReferenceInSceneFile, currentRound, false, foundUnitInList);
				if(*foundUnitInList)
				{
					foundUnit = tempUnit;
				}
			}
		}


		if((currentUnitInList->next == NULL) && (!(*foundUnitInList)) && (topLevelInTree))
		{
			foundUnit = currentUnitInList;

			currentUnitInList->name = unitReferenceInSceneFile->name;
			currentUnitInList->team = unitReferenceInSceneFile->colour;	//not currently used, as there is 1 UnitListClass created for every team
			currentUnitInList->unitDetails = unitReferenceInSceneFile->subModelDetails;
			currentUnitInList->status = true;
			currentUnitInList->roundSpawned = currentRound;

			UnitListClass * newUnitList = new UnitListClass();
			currentUnitInList->next = newUnitList;


		}

		currentUnitInList = currentUnitInList->next;

	}

	return foundUnit;
}





/*
//not used
void addExperienceToPlayer(Experience * experience, Player * player)
{
	player->currentExperience = experience;

	Experience * newExperience = new Experience();
	player->currentExperience->next = newExperience;
	player->currentExperience = player->currentExperience->next;
}

//not used
void addExperienceToUnit(Experience * experience, ModelDetails * unit)
{
	unit->currentExperience = experience;

	Experience * newExperience = new Experience();
	unit->currentExperience->next = newExperience;
	unit->currentExperience = unit->currentExperience->next;
}

//not used
void addExperienceInputToExperience(ExperienceInput * experienceInput, Experience * experience)
{
	experience->currentExperienceInput = experienceInput;

	ExperienceInput * newExperienceInput = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput;
	currentExperienceInput = currentExperienceInput->next;
}
*/




//addExperienceFromUnitCombatDecision(unit, unitOpponent, unitCombatType, combatExperienceResult);



bool addExperiencesFromUnitDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList, Reference * initialReferenceInThisPhaseStartScene)
{
	bool result = true;
	
	if(!addExperienceFromUnitPropertiesDecision(unitReference, unitReferenceOpponent, unitDecision, firstUnitInUnitList))		//arbitrary player, OLD=defenderPlayer
	{
		result = false;
		cout << "Error: cannot addExperienceFromUnitPropertiesDecision" << endl;
	}
	if(!addExperienceFromUnitCombatDecision(unitReference, unitReferenceOpponent, unitDecision, firstUnitInUnitList))		//arbitrary player, OLD=defenderPlayer
	{
		result = false;
		cout << "Error: cannot addExperienceFromUnitCombatDecision" << endl;
	}
	if(!addExperienceFromUnitGlobalDecision(initialReferenceInThisPhaseStartScene, unitReference, unitReferenceOpponent, firstUnitInUnitList))
	{
		result = false;
		cout << "Error: cannot addExperienceFromUnitGlobalDecision" << endl;
	}
	
	return result;
}

bool addExperienceFromUnitPropertiesDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList)
{
	bool result = true;

	bool unitIDFound = false;

	//cout << "aa1" << endl;
	//cout << "firstUnitInUnitList->name = " << firstUnitInUnitList->name << endl;
	//cout << "firstUnitInUnitList->next->name = " << firstUnitInUnitList->next->name << endl;
	//cout << "unitReference->colour" << unitReference->colour << endl;
	//cout << "firstUnitInUnitList->name" << firstUnitInUnitList->name << endl;
	UnitListClass * unitInUnitList = searchUnitListFindUnit(firstUnitInUnitList, unitReference->name, unitReference->colour, &unitIDFound);
	//cout << "aa2" << endl;
	if(unitIDFound)
	{
		Experience * currentExperience = unitInUnitList->currentExperience[GAME_INDEX_OF_PROPERTIES_EXPERIENCE_NN];

		generateExperienceFromUnitPropertiesDecision(unitReference, unitReferenceOpponent, unitDecision, currentExperience);

		Experience * newExperience = new Experience();
		unitInUnitList->currentExperience[GAME_INDEX_OF_PROPERTIES_EXPERIENCE_NN]->next = newExperience;
		unitInUnitList->currentExperience[GAME_INDEX_OF_PROPERTIES_EXPERIENCE_NN] = unitInUnitList->currentExperience[GAME_INDEX_OF_PROPERTIES_EXPERIENCE_NN]->next;

	}
	else
	{
		cout << "Error: unit not found in unit list" << endl;
		result = false;
	}


	return result;
}

bool addExperienceFromUnitCombatDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList)
{
	bool result = true;

	bool unitIDFound = false;

	//cout << "aa1" << endl;
	//cout << "firstUnitInUnitList->name = " << firstUnitInUnitList->name << endl;
	//cout << "firstUnitInUnitList->next->name = " << firstUnitInUnitList->next->name << endl;
	//cout << "unitReference->colour" << unitReference->colour << endl;
	//cout << "firstUnitInUnitList->name" << firstUnitInUnitList->name << endl;
	UnitListClass * unitInUnitList = searchUnitListFindUnit(firstUnitInUnitList, unitReference->name, unitReference->colour, &unitIDFound);
	//cout << "aa2" << endl;
	if(unitIDFound)
	{
		Experience * currentExperience = unitInUnitList->currentExperience[GAME_INDEX_OF_COMBAT_EXPERIENCE_NN];

		generateExperienceFromUnitCombatDecision(unitReference, unitReferenceOpponent, unitDecision, currentExperience);

		Experience * newExperience = new Experience();
		unitInUnitList->currentExperience[GAME_INDEX_OF_COMBAT_EXPERIENCE_NN]->next = newExperience;
		unitInUnitList->currentExperience[GAME_INDEX_OF_COMBAT_EXPERIENCE_NN] = unitInUnitList->currentExperience[GAME_INDEX_OF_COMBAT_EXPERIENCE_NN]->next;

	}
	else
	{
		cout << "Error: unit not found in unit list" << endl;
		result = false;
	}


	return result;
}


void generateExperienceFromUnitPropertiesDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, Experience * currentExperience)
{

	/*
	now store every property of the parties as experience inputs:

	define experience inputs for a unit properties decision experience:

	unit1 recordOfUnitTypeDetails
	unit1 recordOfUnitCombatDetailsDefenceHead
	unit1 recordOfUnitCombatDetailsDefenceTorso
	unit1 recordOfUnitCombatDetailsDefenceShield
	unit1 recordOfUnitCombatDetailsAttackCloseCombat
	unit1 recordOfUnitCombatDetailsAttackLongDistance
	unit2 recordOfUnitTypeDetails
	unit2 recordOfUnitCombatDetailsDefenceHead
	unit2 recordOfUnitCombatDetailsDefenceTorso
	unit2 recordOfUnitCombatDetailsDefenceShield
	unit2 recordOfUnitCombatDetailsAttackCloseCombat
	unit2 recordOfUnitCombatDetailsAttackLongDistance
	distance between two units

		only assume the AI Agent knows the properties of the unit.
		do not assume the AI Agent knows the rules of the game; ie do not save high level information Eg attackClass, defenceClass, etc
	*/

	//cout << "3d1" << endl;

	currentExperience->classTargetValue = unitDecision;
	ExperienceInput * currentExperienceInput = currentExperience->firstExperienceInput;

	ModelDetails * unitDetails = unitReference->subModelDetails;
	ModelDetails * unitDetailsOpponent = unitReferenceOpponent->subModelDetails;

	RecordClass * currentRecord;

	//cout << "3d2" << endl;


	currentRecord = unitDetails->recordOfUnitTypeDetails;

	//cout << "unitReference->name = " << unitReference->name << endl;
	//cout << "unitReferenceOpponent->name = " << unitReferenceOpponent->name << endl;

	while(currentRecord->next != NULL)
	{
		//cout << "3d2a1" << endl;

		//cout << "currentRecord->numberOfThisPartIDInTheUnit = " << (double)(currentRecord->numberOfThisPartIDInTheUnit) << endl;
		currentExperienceInput->inputValue = normaliseExperienceInput((double)(currentRecord->numberOfThisPartIDInTheUnit), COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		//cout << "3d2a2" << endl;

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		//cout << "3d2a3" << endl;

		currentRecord = currentRecord->next;
	}

	//cout << "3d2a" << endl;

	currentRecord = unitDetails->recordOfUnitCombatDetailsDefenceHead;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}
	
	currentRecord = unitDetails->recordOfUnitCombatDetailsDefenceTorso;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}
	
	currentRecord = unitDetails->recordOfUnitCombatDetailsDefenceShield;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}
	
	currentRecord = unitDetails->recordOfUnitCombatDetailsAttackCloseCombat;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	currentRecord = unitDetails->recordOfUnitCombatDetailsAttackLongDistance;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	currentRecord = unitDetailsOpponent->recordOfUnitTypeDetails;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	currentRecord = unitDetailsOpponent->recordOfUnitCombatDetailsDefenceHead;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}
	currentRecord = unitDetailsOpponent->recordOfUnitCombatDetailsDefenceTorso;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	currentRecord = unitDetailsOpponent->recordOfUnitCombatDetailsDefenceShield;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}
	
	currentRecord = unitDetailsOpponent->recordOfUnitCombatDetailsAttackCloseCombat;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	currentRecord = unitDetailsOpponent->recordOfUnitCombatDetailsAttackLongDistance;
	while(currentRecord->next != NULL)
	{
		currentExperienceInput->inputValue = normaliseExperienceInput((double)currentRecord->numberOfThisPartIDInTheUnit, COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);

		ExperienceInput * newExperienceInput = new ExperienceInput();
		currentExperienceInput->next = newExperienceInput;
		currentExperienceInput = currentExperienceInput->next;

		currentRecord = currentRecord->next;
	}

	//cout << "3d3" << endl;

	//cout << "3d4" << endl;

	/*
	currentExperienceInput->inputValue = normaliseExperienceInput((double)combatExperienceResult, COMBAT_EXPERIENCE_INPUT_RESULT_MAX);
	ExperienceInput * newExperienceInput2 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput2;
	currentExperienceInput = currentExperienceInput->next;
	*/

		int maximumMovementDistanceForAnyUnit = 0;
		RulesClass * currentReferenceRulesClass = LRRCrulesUnitTypeCatagories;
		while(currentReferenceRulesClass->next != NULL)
		{
			maximumMovementDistanceForAnyUnit = maxInt(maximumMovementDistanceForAnyUnit, currentReferenceRulesClass->attribute4);

			currentReferenceRulesClass = currentReferenceRulesClass->next;
		}
		//cout << "DEBUG: maximumMovementDistanceForAnyUnit = " << maximumMovementDistanceForAnyUnit << endl;

	//cout << "3d5" << endl;

	double distanceBetweenUnits = calculateTheDistanceBetweenTwoUnits(&(unitReference->absolutePosition), &(unitReferenceOpponent->absolutePosition));
	currentExperienceInput->inputValue = normaliseExperienceInput(distanceBetweenUnits, maximumMovementDistanceForAnyUnit);
	ExperienceInput * newExperienceInput3 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput3;
	currentExperienceInput = currentExperienceInput->next;
	//add 2D relationship map information here - can be added to a hypbrid 1D/2D neural network...
	
	
	
	//cout << "3d6" << endl;
}

void generateExperienceFromUnitCombatDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, Experience * currentExperience)
{

	/*
	now store every combat level of the parties as experience inputs:

	define experience inputs for a unit combat decision experience:

	unit1 movementSpeed
	unit1 closeCombatAttackTotal
	unit1 longDistanceAttackTotal
	unit1 defenceTotal
	unit2 movementSpeed
	unit2 closeCombatAttackTotal
	unit2 longDistanceAttackTotal
	unit2 defenceTotal
	distance between two units

		only assume the AI Agent knows the properties of the unit.
		do not assume the AI Agent knows the rules of the game; ie do not save high level information Eg attackClass, defenceClass, etc
	*/




	int maximumMovementDistanceForAnyUnit = 0;
	int maximumCloseCombatAttackTotalForAnyUnit = 0;
	int maximumLongDistanceAttackTotalForAnyUnit = 0;
	int maximumDefenceHeadTotalForAnyUnit = 0;
	int maximumDefenceTorsoTotalForAnyUnit = 0;
	int maximumDefenceShieldTotalForAnyUnit = 0;
	int maximumDefenceTotalForAnyUnit = 0;
	RulesClass * currentReferenceRulesClass;
	
	currentReferenceRulesClass = LRRCrulesUnitTypeCatagories;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumMovementDistanceForAnyUnit = maxInt(maximumMovementDistanceForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}

	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsAttackCloseCombat;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumCloseCombatAttackTotalForAnyUnit = maxInt(maximumCloseCombatAttackTotalForAnyUnit, (currentReferenceRulesClass->attribute4 + currentReferenceRulesClass->attribute5));

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsAttackLongDistance;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumLongDistanceAttackTotalForAnyUnit = maxInt(maximumLongDistanceAttackTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceHead;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceHeadTotalForAnyUnit = maxInt(maximumDefenceHeadTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceTorso;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceTorsoTotalForAnyUnit = maxInt(maximumDefenceTorsoTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceShield;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceShieldTotalForAnyUnit = maxInt(maximumDefenceShieldTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	maximumDefenceTotalForAnyUnit = maximumDefenceHeadTotalForAnyUnit + maximumDefenceTorsoTotalForAnyUnit + maximumDefenceShieldTotalForAnyUnit;
	
	
	
	//cout << "3d1" << endl;

	currentExperience->classTargetValue = unitDecision;
	ExperienceInput * currentExperienceInput = currentExperience->firstExperienceInput;

	ModelDetails * unitDetails = unitReference->subModelDetails;
	ModelDetails * unitDetailsOpponent = unitReferenceOpponent->subModelDetails;
	
	//fill experience inputs
	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetails->movementSpeed, maximumMovementDistanceForAnyUnit);
	ExperienceInput * newExperienceInput1 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput1;
	currentExperienceInput = currentExperienceInput->next;
		
	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetails->closeCombatAttackTotal, maximumCloseCombatAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput2 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput2;
	currentExperienceInput = currentExperienceInput->next;

	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetails->longDistanceAttackTotal, maximumLongDistanceAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput3 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput3;
	currentExperienceInput = currentExperienceInput->next;

	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetails->defenceTotal, maximumDefenceTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput4 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput4;
	currentExperienceInput = currentExperienceInput->next;
	
					
	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetailsOpponent->movementSpeed, maximumMovementDistanceForAnyUnit);
	ExperienceInput * newExperienceInput5 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput5;
	currentExperienceInput = currentExperienceInput->next;
		
	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetailsOpponent->closeCombatAttackTotal, maximumCloseCombatAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput6 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput6;
	currentExperienceInput = currentExperienceInput->next;

	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetailsOpponent->longDistanceAttackTotal, maximumLongDistanceAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput7 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput7;
	currentExperienceInput = currentExperienceInput->next;

	currentExperienceInput->inputValue = normaliseExperienceInput(unitDetailsOpponent->defenceTotal, maximumDefenceTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
	ExperienceInput * newExperienceInput8 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput8;
	currentExperienceInput = currentExperienceInput->next;
						
					

	//cout << "3d5" << endl;

	double distanceBetweenUnits = calculateTheDistanceBetweenTwoUnits(&(unitReference->absolutePosition), &(unitReferenceOpponent->absolutePosition));
	currentExperienceInput->inputValue = normaliseExperienceInput(distanceBetweenUnits, maximumMovementDistanceForAnyUnit);
	ExperienceInput * newExperienceInput9 = new ExperienceInput();
	currentExperienceInput->next = newExperienceInput9;
	currentExperienceInput = currentExperienceInput->next;
	//add 2D relationship map information here - can be added to a hypbrid 1D/2D neural network...
	
	
	
	//cout << "3d6" << endl;
}





bool addExperienceFromUnitGlobalDecision(Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Reference * unitReferenceOpponent, UnitListClass * firstUnitInUnitList)
{
	bool result = true;

	bool unitIDFound = false;

	//cout << "aa1" << endl;
	//cout << "firstUnitInUnitList->name = " << firstUnitInUnitList->name << endl;
	//cout << "firstUnitInUnitList->next->name = " << firstUnitInUnitList->next->name << endl;
	//cout << "unitReference->colour" << unitReference->colour << endl;
	//cout << "firstUnitInUnitList->name" << firstUnitInUnitList->name << endl;
	UnitListClass * unitInUnitList = searchUnitListFindUnit(firstUnitInUnitList, unitReference->name, unitReference->colour, &unitIDFound);
	//cout << "aa2" << endl;
	if(unitIDFound)
	{
		Experience * currentExperience = unitInUnitList->currentExperience[GAME_INDEX_OF_GLOBAL_EXPERIENCE_NN];

		generateExperienceFromGlobalDecision(firstUnitInUnitList, initialReferenceInThisPhaseStartScene, unitReference, unitReferenceOpponent, currentExperience);

		Experience * newExperience = new Experience();
		unitInUnitList->currentExperience[GAME_INDEX_OF_GLOBAL_EXPERIENCE_NN]->next = newExperience;
		unitInUnitList->currentExperience[GAME_INDEX_OF_GLOBAL_EXPERIENCE_NN] = unitInUnitList->currentExperience[GAME_INDEX_OF_GLOBAL_EXPERIENCE_NN]->next;

	}
	else
	{
		cout << "Error: unit not found in unit list" << endl;
		result = false;
	}


	return result;
}

bool generateExperienceFromGlobalDecision(UnitListClass * firstUnitInUnitList, Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Reference * unitReferenceOpponent, Experience * currentExperience)
{
	bool result = true;
	
	
	int xPosRelToUnit = (unitReferenceOpponent->absolutePosition.x - unitReference->absolutePosition.x);
	int yPosRelToUnit = (unitReferenceOpponent->absolutePosition.y - unitReference->absolutePosition.y);
	if((abs(xPosRelToUnit) > GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS) || (abs(yPosRelToUnit) > GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS))
	{
		xPosRelToUnit = GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS;
		yPosRelToUnit = GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS;
	}
	int xPosRelToUnitAsTargetValue = xPosRelToUnit + GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS;
	int yPosRelToUnitAsTargetValue = yPosRelToUnit + GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS;
	currentExperience->classTargetValue = xPosRelToUnitAsTargetValue*GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D + yPosRelToUnitAsTargetValue;

					
					
	//fill experience input tree with blank input values
	ExperienceInput * currentExperienceInput = currentExperience->firstExperienceInput;
	
	for(int x=-GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; x<=GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; x++)
	{
		for(int y=-GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; y<=GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; y++)
		{
			for(int t=0; t<GLOBAL_EXPERIENCE_NUM_INPUT_TYPES; t++)
			{
				ExperienceInput * newExperienceInput = new ExperienceInput();
				currentExperienceInput->next = newExperienceInput;
				currentExperienceInput = currentExperienceInput->next;
			}
		}	
	}
	
	//generate2DMemoryMapExperience of LDAttack, CCattack, defence and movement totals [4 input types] for each unit group centred around unitReference - linear/Euclidian spatial map going from 0 to 100 radius in 10 unit intervals - total inputs = 10x10x4
	
	if(!generate2DMemoryMapExperience(firstUnitInUnitList, initialReferenceInThisPhaseStartScene, unitReference, currentExperience))
	{
		result = false;
	}
	return result;

	//cout << "3d6" << endl;
}




bool generate2DMemoryMapExperience(UnitListClass * firstUnitInGroup, Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Experience * currentExperience)
{
	bool result = true;
	UnitListClass * currentUnitInGroup = firstUnitInGroup;
	
	int maximumMovementDistanceForAnyUnit = 0;
	int maximumCloseCombatAttackTotalForAnyUnit = 0;
	int maximumLongDistanceAttackTotalForAnyUnit = 0;
	int maximumDefenceHeadTotalForAnyUnit = 0;
	int maximumDefenceTorsoTotalForAnyUnit = 0;
	int maximumDefenceShieldTotalForAnyUnit = 0;
	int maximumDefenceTotalForAnyUnit = 0;
	RulesClass * currentReferenceRulesClass;
	
	currentReferenceRulesClass = LRRCrulesUnitTypeCatagories;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumMovementDistanceForAnyUnit = maxInt(maximumMovementDistanceForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}

	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsAttackCloseCombat;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumCloseCombatAttackTotalForAnyUnit = maxInt(maximumCloseCombatAttackTotalForAnyUnit, (currentReferenceRulesClass->attribute4 + currentReferenceRulesClass->attribute5));

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsAttackLongDistance;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumLongDistanceAttackTotalForAnyUnit = maxInt(maximumLongDistanceAttackTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceHead;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceHeadTotalForAnyUnit = maxInt(maximumDefenceHeadTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceTorso;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceTorsoTotalForAnyUnit = maxInt(maximumDefenceTorsoTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	currentReferenceRulesClass = LRRCrulesUnitCombatDetailsDefenceShield;
	while(currentReferenceRulesClass->next != NULL)
	{
		maximumDefenceShieldTotalForAnyUnit = maxInt(maximumDefenceShieldTotalForAnyUnit, currentReferenceRulesClass->attribute4);

		currentReferenceRulesClass = currentReferenceRulesClass->next;
	}
	maximumDefenceTotalForAnyUnit = maximumDefenceHeadTotalForAnyUnit + maximumDefenceTorsoTotalForAnyUnit + maximumDefenceShieldTotalForAnyUnit;
	
	
	
	while(currentUnitInGroup->next != NULL)
	{
		if(currentUnitInGroup->status == true)
		{
			//1. create a dummy reference for details calculations
			Reference * currentReferenceInGroup = new Reference(true);
			currentReferenceInGroup->colour = currentUnitInGroup->team;
			currentReferenceInGroup->name = currentUnitInGroup->name;
			if(!searchSceneReferenceListAndDetermineTheDetailsOfAParticularUnitSubmodel(currentReferenceInGroup, initialReferenceInThisPhaseStartScene, NULL, false))
			{
				cout << "generate2DMemoryMapExperience(): Error - unit reference not found in reference list" << endl;
				cout << "\terror: cannot find player unit reference in scene file, name = " << currentReferenceInGroup->name << " id=" << currentReferenceInGroup->colour << endl;
				result = false;
			}
			else
			{
				performFinalUnitGroupClassCalculations(currentReferenceInGroup->subModelDetails);

				int xPosRelToUnit = (currentReferenceInGroup->absolutePosition.x - unitReference->absolutePosition.x);
				int yPosRelToUnit = (currentReferenceInGroup->absolutePosition.y - unitReference->absolutePosition.y);
				bool refIn2DMemoryMap = true;
				if((abs(xPosRelToUnit) > GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS) || (abs(yPosRelToUnit) > GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS))
				{
					refIn2DMemoryMap = false;
				}
				if(refIn2DMemoryMap)
				{
					ExperienceInput * currentReferenceExperienceInput = findExperienceInputIn2DMemoryMap(xPosRelToUnit, yPosRelToUnit, 0, currentExperience);
					
					//fill experience inputs
					currentReferenceExperienceInput->inputValue = normaliseExperienceInput(currentReferenceInGroup->subModelDetails->movementSpeed, maximumMovementDistanceForAnyUnit);
					currentReferenceExperienceInput = currentReferenceExperienceInput->next;

					currentReferenceExperienceInput->inputValue = normaliseExperienceInput(currentReferenceInGroup->subModelDetails->closeCombatAttackTotal, maximumCloseCombatAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
					currentReferenceExperienceInput = currentReferenceExperienceInput->next;
	
					currentReferenceExperienceInput->inputValue = normaliseExperienceInput(currentReferenceInGroup->subModelDetails->longDistanceAttackTotal, maximumLongDistanceAttackTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
					currentReferenceExperienceInput = currentReferenceExperienceInput->next;
					
					currentReferenceExperienceInput->inputValue = normaliseExperienceInput(currentReferenceInGroup->subModelDetails->defenceTotal, maximumDefenceTotalForAnyUnit*COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX);
					currentReferenceExperienceInput = currentReferenceExperienceInput->next;	
				}
				
				if(currentUnitInGroup->isUnitGroup)
				{
					generate2DMemoryMapExperience(currentUnitInGroup->firstUnitInUnitGroup, initialReferenceInThisPhaseStartScene, unitReference, currentExperience);
				}
			}
			
			delete currentReferenceInGroup;

		}

		currentUnitInGroup = currentUnitInGroup->next;
	}
	return result;	
}





ExperienceInput * findExperienceInputIn2DMemoryMap(int xPos, int yPos, int tPos, Experience * currentExperience)
{
	ExperienceInput * currentExperienceInput = currentExperience->firstExperienceInput;
	for(int x=-GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; x<xPos; x++)
	{
		for(int y=-GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; y<GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; y++)
		{
			for(int t=0; t<GLOBAL_EXPERIENCE_NUM_INPUT_TYPES; t++)
			{
				currentExperienceInput = currentExperienceInput->next;
			}
		}	
	}
	for(int y=-GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS; y<yPos; y++)
	{
		for(int t=0; t<GLOBAL_EXPERIENCE_NUM_INPUT_TYPES; t++)
		{
			currentExperienceInput = currentExperienceInput->next;
		}
	}
	for(int t=0; t<tPos; t++)
	{
		currentExperienceInput = currentExperienceInput->next;
	}
	return currentExperienceInput;
}



/*
void generate2DMemoryMapExperience(Reference * firstReferenceInLayer,  Reference * unitReference, Experience * currentExperience)
{
	Reference * currentReference = firstReferenceInLayer;
	
	while(currentReference->next != NULL)
	{
		if(currentReference->isSubModelReference)
		{
			int xPosRelToUnit = (currentReference->absolutePosition.x - unitReference->absolutePosition.x);
			int yPosRelToUnit = (currentReference->absolutePosition.y - unitReference->absolutePosition.y);
			bool refIn2DMemoryMap = true;
			if((xPosRelToUnit > EXPERIENCE_FROM_UNIT_TARGETTING_OUTCOME_RADIUS/EXPERIENCE_FROM_UNIT_TARGETTING_OUTCOME_INTERVAL) || (yPosRelToUnit > EXPERIENCE_FROM_UNIT_TARGETTING_OUTCOME_RADIUS/EXPERIENCE_FROM_UNIT_TARGETTING_OUTCOME_INTERVAL))
			{
				refIn2DMemoryMap = false;
			}
			if(refIn2DMemoryMap)
			{
				ExperienceInput * currentReferenceExperienceInput = findExperienceInputIn2DMemoryMap(xPosRelToUnit, yPosRelToUnit, 0, currentExperience);
			}
			
			generate2DMemoryMapExperience(currentReference->firstReferenceWithinSubModel, unitReference, currentExperience)
		}
		
		currentReference = currentReference->next;
		
	}	
}
*/


double normaliseExperienceInput(double currentInputValue, double maxInputValue)
{
	double minInputValue = 0.0;
	if(currentInputValue > maxInputValue)
	{//deal with redundant situations, eg a single unit has two swords - nb only one part type can be used per unit,  and so the following is a valid correction;
		currentInputValue = maxInputValue;
	}

	double normalisedExperienceInput;

	if(compareDoubles(maxInputValue, minInputValue))
	{
		/*
		cout << "currentInputValue = " << currentInputValue << endl;
		cout << "minInputValue = " << minInputValue << endl;
		cout << "maxInputValue = " << maxInputValue << endl;
		cout << "a" << endl;
		*/
		normalisedExperienceInput = currentInputValue;
	}
	else
	{
		/*
		cout << "currentInputValue = " << currentInputValue << endl;
		cout << "minInputValue = " << minInputValue << endl;
		cout << "maxInputValue = " << maxInputValue << endl;
		cout << "b" << endl;
		*/
		normalisedExperienceInput = ((currentInputValue - minInputValue) / (maxInputValue - minInputValue));
	}
	return normalisedExperienceInput;
}



















void updatePlayerStatus(Player * initialPlayerInList)
{
	Player * currentPlayer = initialPlayerInList;

	while(currentPlayer->next != NULL)
	{
		if(!searchUnitListAndCheckThatSomePlayerUnitsAreAlive(currentPlayer->id, initialPlayerInList->firstUnitInUnitList))		//arbitrary player, OLD=defenderPlayer
		{
			currentPlayer->status = false;
		}

		currentPlayer = currentPlayer->next;
	}

}



bool searchUnitListAndCheckThatSomePlayerUnitsAreAlive(int playerTeam, UnitListClass * firstUnitInUnitGroup)
{
	bool playerStillHasAUnitAlive = false;

	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		if((currentUnitInList->team == playerTeam) && (currentUnitInList->status == ALIVE))
		{
			playerStillHasAUnitAlive = true;
		}

		if(currentUnitInList->isUnitGroup)
		{
			if(searchUnitListAndCheckThatSomePlayerUnitsAreAlive(playerTeam, currentUnitInList->firstUnitInUnitGroup))
			{
				playerStillHasAUnitAlive = true;
			}
		}

		currentUnitInList = currentUnitInList->next;
	}

	return playerStillHasAUnitAlive;

}









void writeExperienceListToFile(char * fileName, Experience * firstExperienceInList)
{
	ofstream experienceDataSetOFStreamObject(fileName);

	Experience * currentExperience = firstExperienceInList;
	while(currentExperience->next != NULL)
	{
		addExperienceToOFStream(&experienceDataSetOFStreamObject, currentExperience);
		currentExperience = currentExperience->next;
	}

	experienceDataSetOFStreamObject.close();
}

void addExperienceToOFStream(ofstream * experienceDataSetOFStreamObject, Experience * experience)
{
	char tempString[MAX_EXPERIENCE_SIGNIFICANT_BITS];

	sprintf(tempString, "%0.6f", experience->classTargetValue);
	string tempStringPlusPlus = tempString;
	for(int i = 0; i<(int)(tempStringPlusPlus.length());i++)
	{
		experienceDataSetOFStreamObject->put(tempString[i]);
	}

	ExperienceInput * currentExperienceInput = experience->firstExperienceInput;
	while(currentExperienceInput->next != NULL)
	{
		sprintf(tempString, "%0.6f",  currentExperienceInput->inputValue);
		string tempStringPlusPlus = tempString;
		for(int i = 0; i<(int)(tempStringPlusPlus.length());i++)
		{
			experienceDataSetOFStreamObject->put(tempString[i]);
		}
		experienceDataSetOFStreamObject->put(CHAR_COMMA);
		currentExperienceInput = currentExperienceInput->next;
	}

	experienceDataSetOFStreamObject->put(CHAR_NEW_LINE);
}


void addAllUnitExperiencesToOFStream(ofstream * experienceDataSetOFStreamObject, UnitListClass * firstUnitInUnitGroup, int nnIndex)
{
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		#ifndef DEBUG_DO_NOT_PERFORM_KILL_RATIO_CHECKS_BEFORE_ADDING_UNIT_EXPERIENCES_TO_NN
		if(checkAverageKillRatioForUnitGroup(currentUnitInList) || (currentUnitInList->unitDetails->numPerson == 1))
		{	
		#endif	
			Experience * currentExperience = firstUnitInUnitGroup->firstExperience[nnIndex];
			while(currentExperience->next != NULL)
			{
				addExperienceToOFStream(experienceDataSetOFStreamObject, currentExperience);
				currentExperience = currentExperience->next;
			}

			if(currentUnitInList->isUnitGroup)
			{
				addAllUnitExperiencesToOFStream(experienceDataSetOFStreamObject, currentUnitInList->firstUnitInUnitGroup, nnIndex);
			}
		#ifndef DEBUG_DO_NOT_PERFORM_KILL_RATIO_CHECKS_BEFORE_ADDING_UNIT_EXPERIENCES_TO_NN
		}
		#endif

		currentUnitInList = currentUnitInList->next;
	}
}

void addPlayerUnitExperiencesToOFStream(ofstream * experienceDataSetOFStreamObject, UnitListClass * firstUnitInUnitGroup, int unitGroupTeam, int nnIndex)
{
	UnitListClass * currentUnitInList = firstUnitInUnitGroup;

	while(currentUnitInList->next != NULL)
	{
		#ifndef DEBUG_DO_NOT_PERFORM_KILL_RATIO_CHECKS_BEFORE_ADDING_UNIT_EXPERIENCES_TO_NN
		if(checkAverageKillRatioForUnitGroup(currentUnitInList) || (currentUnitInList->unitDetails->numPerson == 1))
		{
		#endif
			if(currentUnitInList->team == unitGroupTeam)
			{
				Experience * currentExperience = firstUnitInUnitGroup->firstExperience[nnIndex];
				while(currentExperience->next != NULL)
				{
					addExperienceToOFStream(experienceDataSetOFStreamObject, currentExperience);
					currentExperience = currentExperience->next;
				}
			}

			if(currentUnitInList->isUnitGroup)
			{
				addPlayerUnitExperiencesToOFStream(experienceDataSetOFStreamObject, currentUnitInList->firstUnitInUnitGroup, unitGroupTeam, nnIndex);
			}
		#ifndef DEBUG_DO_NOT_PERFORM_KILL_RATIO_CHECKS_BEFORE_ADDING_UNIT_EXPERIENCES_TO_NN
		}
		#endif
		
		currentUnitInList = currentUnitInList->next;
	}
}





/*OLD:


void addExperienceToByteArray()
{

}

	long ExperienceDataSetByteArrayCurrentPosition = 0;
	long ExperienceDataSetByteArraySize = XML_FILE_MAX_SIZE;
	char * ExperienceDataSetByteArray = new char[EXPERIENCE_DATASET_MAX_SIZE];

*/
