/*******************************************************************************
 *
 * File Name: LRRCgameAI.cpp
 * Author: Richard Baxter
 * Project: Lego Rules CG Rounds Checker
 * Project Version: 1l3j 2-Sept-08
 * Project First Internal Release: 1aXx 18-Sept-05 (C)
 * Project Second Internal Release: 2aXx 02-April-06 (convert to C++)
 * Project Third Internal Release: 2b7d 26-Sept-06 (added sprites)
 * Project Fourth Internal Release: 2c4a 25-Feb-07 [rotation of models supported, copyCombatRelevantInheritedImmediateParentUnitDetailsIntoChildObject() no longer used]
 * Project Fifth Internal Release: 2d11a 3-June-07
 *
 * NB phase specific sprites are yet to be added to scene files based upon movement/combat outcomes
 * NB phase specific sprites are yet to be added to scene files based upon future movement/combat options
 *******************************************************************************/

#ifndef HEADER_LRRC_GAME_AI
#define HEADER_LRRC_GAME_AI

#include "LDglobalDefs.h"
//#assert USE_ANN

#include "ANNexperienceClass.h"
#include "LRRCplayerClass.h"
#include "LRRCunitClass.h"
#include "LDreferenceClass.h"

#define COMBAT_TYPE_CLOSE_COMBAT (1)
#define COMBAT_TYPE_LONG_DISTANCE (2)

#define ALIVE (true)
#define DEAD (false)





#define COMBAT_EXPERIENCE_OUTPUT_DECISION_IGNORE 0				//ignore target and continue moving towards goal (/ not worth attacking and not worth evading opponent)
#define COMBAT_EXPERIENCE_OUTPUT_DECISION_ATTACK_CLOSECOMBAT 1		//decided to attack {or move to attack} opponent as it is considered easy (/ profitable for the mission)
#define COMBAT_EXPERIENCE_OUTPUT_DECISION_ATTACK_LONGDISTANCE 2				//decided to move closer to the unit [possibly for close combat/long range attack]
#define COMBAT_EXPERIENCE_OUTPUT_DECISION_EVADE 3				//decide not to attack {or move to attack} opponent - being attacked will result a serious danger (/ danger to the mission) and therefore must be evaded at the cost of a time delay
#define COMBAT_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS 4
	//COMBAT_EXPERIENCE_OUTPUT_DECISION_IGNORE and COMBAT_EXPERIENCE_OUTPUT_DECISION_EVADE may be merged in the future...
	
//OLD: #define COMBAT_EXPERIENCE_OUTPUT_DECISION_MOVE_TO_ATTACK 3
//OLD: #define COMBAT_EXPERIENCE_OUTPUT_DECISION_ATTACK_LONGDISTANCE 4

#ifdef DEBUG_HAVNT_YET_STARTED_TESTING_UNITGROUPS
#define MAXIMUM_UNIT_GROUP_SIZE (1)
#else
#define MAXIMUM_UNIT_GROUP_SIZE (10)
#endif
#define COMBAT_EXPERIENCE_NUMBER_OF_A_PART_IN_A_UNIT_MAX (MAXIMUM_UNIT_GROUP_SIZE)
#define COMBAT_EXPERIENCE_MAXIMUM_LEVEL (6)

#define GLOBAL_EXPERIENCE_RADIUS (50)
#define GLOBAL_EXPERIENCE_INTERVAL (25)
#define GLOBAL_EXPERIENCE_NUM_INPUT_TYPES (4)
#define GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS (GLOBAL_EXPERIENCE_RADIUS/GLOBAL_EXPERIENCE_INTERVAL)
#define GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D (((GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_RADIUS)*2)+1)
#define GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS (GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D*GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D)
#define GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_INPUT_DECISIONS (GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D*GLOBAL_EXPERIENCE_NUMBER_OF_DIFFERENT_OUTPUT_DECISIONS_1D*GLOBAL_EXPERIENCE_NUM_INPUT_TYPES)

#define GAME_UNIT_POINTS_LONG_DISTANCE_RANGE_MODIFIER (10)	
	//eg a long bow with range 20, and long distance attack value d4 [level 2], recieves (20/10=2 + 2) = 4 points. 
	//eg a crossbow with range 15, and long distance attack value d6 [level 3], recieves (15/10=1 + 3) = 4 points
	//eg a spearman with range 10, and long distance attack value d8 [level 3], recieves (10/10=1 + 3) = 4 points
#define GAME_UNIT_POINTS_MOVEMENT_RANGE_MODIFIER (5)
	//eg a knight receives (20) 20/5 = 4 points
	//eg a footsoldier (10) receives 10/5 = 2 points

#define GAME_AI_NN_EXPERIENCE_RECORD_MIN_KILL_RATIO_REQUIREMENT (1.0)
	
/*
//not used
void addExperienceToPlayer(Experience * experience, Player * player);
void addExperienceToUnit(Experience * experience, ModelDetails * unit);
void addExperienceInputToExperience(ExperienceInput * experienceInput, Experience * experience);
*/


void parseSceneFileAndFillUnitLists(char sceneFileName[], UnitListClass * firstUnitInUnitList, int currentRound);
	void fillUnitList(Reference * currentReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound);

void parseSceneFileAndUpdateUnitList(char sceneFileName[], UnitListClass * firstUnitInUnitList, int currentRound);
	void updateUnitList(Reference * initialReferenceInSceneFile, UnitListClass * firstUnitInUnitList, int currentRound);	//required after every round
		void updateUnitListWithNewUnits(Reference * currentReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound);
			UnitListClass * searchUnitListForUnitAndIfNotThereAddToList(UnitListClass * firstUnitInUnitGroup, Reference * unitReferenceInSceneFile, int currentRound, bool topLevelInTree, bool * foundUnitInList);
		void updateUnitListWithDeadUnits(Reference * initialReferenceInSceneFile, UnitListClass * firstUnitInUnitGroup, int currentRound);

double normaliseExperienceInput(double currentInputValue, double maxInputValue);


bool determineIfUnitGroupHasAliveUnits(UnitListClass * firstUnitInUnitGroup);
bool checkAverageKillRatioForUnitGroup(UnitListClass * firstUnitInUnitGroup);
	void determineAverageKillRatioForUnitGroup(UnitListClass * firstUnitInUnitGroup, int * killPoints, int * deathPoints);
		int determineUnitWorthInPoints(ModelDetails * unitDetails);

bool addExperiencesFromUnitDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList, Reference * initialReferenceInThisPhaseStartScene);
	bool addExperienceFromUnitPropertiesDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList);
		void generateExperienceFromUnitPropertiesDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, Experience * currentExperience);
	bool addExperienceFromUnitCombatDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, UnitListClass * firstUnitInUnitList);
		void generateExperienceFromUnitCombatDecision(Reference * unitReference, Reference * unitReferenceOpponent, long unitDecision, Experience * currentExperience);
	bool addExperienceFromUnitGlobalDecision(Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Reference * unitReferenceOpponent, UnitListClass * firstUnitInUnitList);
		bool generateExperienceFromGlobalDecision(UnitListClass * firstUnitInUnitList, Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Reference * unitReferenceOpponent, Experience * currentExperience);
			bool generate2DMemoryMapExperience(UnitListClass * firstUnitInGroup, Reference * initialReferenceInThisPhaseStartScene, Reference * unitReference, Experience * currentExperience);
				ExperienceInput * findExperienceInputIn2DMemoryMap(int xPos, int yPos, int tPos, Experience * currentExperience);



void updatePlayerStatus(Player * initialPlayerInList);
	bool searchUnitListAndCheckThatSomePlayerUnitsAreAlive(int playerTeam, UnitListClass * firstUnitInUnitGroup);

void writeExperienceListToFile(char * fileName, Experience * firstExperienceInList);
void addAllUnitExperiencesToOFStream(ofstream * experienceDataSetOFStreamObject, UnitListClass * firstUnitInUnitGroup, int nnIndex);
void addPlayerUnitExperiencesToOFStream(ofstream * experienceDataSetOFStreamObject, UnitListClass * firstUnitInUnitGroup, int unitGroupTeam, int nnIndex);
	void addExperienceToOFStream(ofstream * experienceDataSetOFStreamObject, Experience * experience);

#endif




